/*
 * File:   main.c - AC100Meas
 * PIC18F26J50 - fosc 8.192MHz
 * Created on 2023/01/23, 13:20
 */

#pragma config WDTEN = OFF  // configration register setting
#pragma config OSC = HSPLL  // HSosc 8.192MHz, PLL ON
#pragma config XINST = OFF
#pragma config PLLDIV = 2   // PLL input is 4.096MHz -> 98.304MHz /2
#pragma config CPUDIV = OSC3_PLL3 // CPU clock divider 1/3 16.384MHz

#include <xc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "font.h"
#include "msp3520.h"
#include "sintable.h"

#define FOSC 8192000 // HS osc frequency
#define FINSTR 2048000 // instruction clock frequency
#define COEF_V_PIC 5.8184  // voltage coefficient
#define COEF_V_PCM1808 5.4743
#define LED LATAbits.LATA5
#define SCK2  PORTCbits.RC0
#define MISO2 PORTCbits.RC1
#define SS2   PORTCbits.RC2

#define F50Hz 50
#define F60Hz 60
#define WAVEFORM_BUFFER_SIZE 1600

// ADコンバータ PCM1808 か PIC内蔵ADCかの選択 どちらかをコメントアウトする
//#define WAVEFORM_SAMPLING() pic_waveform_sampling()
#define WAVEFORM_SAMPLING() pcm1808_waveform_sampling()

#define ADCPIC 0
#define ADCPCM1808 1

unsigned int waveform[ WAVEFORM_BUFFER_SIZE ];
int pwf; // waveform pointer
int adc;

unsigned int freq_resion = F60Hz;

int tmr0_c = 0;

// 低優先度割り込み処理 インターバルタイマ
void __interrupt( low_priority )
isr_l( void )
{
// TMR0 interrupt - 31.25Hz interval interrupt
  if( INTCONbits.T0IF ) { 
    LED = 1;
    INTCONbits.T0IF = 0;
    if( tmr0_c  )  tmr0_c --; // 汎用タイマが０でなければデクリメント
    LED = 0;
  }
}

unsigned int plc, plc_avg, last_p;  // PowerLine cycle period counter(raw)
unsigned long int plc_sum; // for averageing
unsigned int plc_n, plc_avg, plc_max, plc_min;
char spi2bytecount;
unsigned char pcmmsb, pcmlsb;

// 高優先度割り込み処理 PCM1808データ受信 電源周期測定 内蔵ADC
// 優先度の高い割り込みから順に処理する
void __interrupt( high_priority )
isr_h( void )
{
  unsigned int p;
// PCM1808 data receive
// PCM1808は最上位バイトから4バイトのデータを送って来る
// 上位2バイト(16ビット)のみを波形メモリに格納する
  if( PIR3bits.SSP2IF ) {
    LED = 1; // 割り込み処理中表示
    PIR3bits.SSP2IF = 0; // 割り込みフラグをクリア
    pcmmsb = SSP2BUF; // 上位バイトを受信
    NOP();
    NOP();
    waveform[pwf] = pcmmsb * 256 + SSP2BUF; // 下位バイトを受信し、波形メモリへ格納
    pwf++; // ポインタ更新
    if( WAVEFORM_BUFFER_SIZE <= pwf ) { PIE3bits.SSP2IE = 0; SSP2CON1bits.SSPEN = 0; }
    // 波形メモリがいっぱいになったらSPIの動作と割り込みを停止する
    p = SSP2BUF; // discard lower data 使用しないデータを破棄(2バイト)
    NOP(); // タイミング合わせ
    NOP();
    NOP();
    NOP();
    p = SSP2BUF; // discard lower data
    PIR3bits.SSP2IF = 0;
    spi2bytecount ++; // デバッグ用に使用した変数
    spi2bytecount &= 0x03;            
    LED = 0; // 割り込み処理終了
  }
// INT0 interrupt - Power line cycle
// TMR1を読みだして前回との差を「周期」として扱う
  if( INTCONbits.INT0F ) { 
    LED = 1; // 割り込み処理中表示
    p = TMR1; // TMR1は2.048MHzでフリーランしている
    INTCONbits.INT0F = 0; // 割り込みフラグをクリア
    plc = p - last_p; // 前回との差を計算
    last_p = p; // 次に備えて保存
    plc_sum += ( unsigned long int ) plc; // 平均をとるため加算
    plc_n ++; // データ取得回数を増加
    if( plc_max < plc ) plc_max = plc; // 最大値
    if( plc_min > plc ) plc_min = plc; // 最小値
    if( 10 == plc_n ) { // データ取得回数が10回になったら、最大値と最小値を捨てて8個分の平均をとる(ノイズ等による誤差を減らす)
      plc_avg = ( unsigned int )( ( plc_sum - ( unsigned long )plc_min - ( unsigned long )plc_max ) / 8L );
      plc_sum = 0; // 平均値用バッファ、回数、最大値、最小値をリセット
      plc_n = 0;
      plc_max = 0, plc_min = 0xFFFF;
    }
    LED = 0; // 割り込み処理終了
  }
// TMR2 interrupt - start PIC ADC TMR2で内蔵ADCをスタートさせる
  if( PIR1bits.TMR2IF ) {
    LED = 1; // 割り込み処理中表示
    ADCON0bits.GO = 1; // サンプリング開始
    PIR1bits.TMR2IF = 0;
    LED = 0; // 割り込み処理終了
  }
// ADC interrupt - collect data 内蔵ADCのデータを波形メモリへ格納
  if( PIR1bits.ADIF ) {
    LED = 1; // 割り込み処理中表示
    waveform[pwf] = ADRESH * 256 + ADRESL; // 2バイトデータを取得、波形メモリへ格納
    pwf++;
    PIR1bits.ADIF = 0; // 割り込みフラグをクリア
    if( WAVEFORM_BUFFER_SIZE <= pwf ) T2CONbits.TMR2ON = 0; // terminate data collection
    LED = 0; // 割り込み処理終了
  }
}

// initialize routine
void
iniz( void )
{
  OSCTUNEbits.PLLEN = 1; // クロックジェネレータのPLLをON
　// クロック周波数はコンフィギュレーションレジスタ(＃pragma）で設定してある

// port / PIN 入出力ピンの設定
  ANCON0 = 0xFE; // RA0/AN0 is analog input
  ANCON1 = 0x1F; // other pins are digital I/O
  UCONbits.USBEN = 0; // USBは使用しない
  UCFGbits.UTRDIS = 1;
  TRISA = 0b00001101; // RA0/AN0 is input, other pins are output
                      // RA2/RA3 are VREF, RA5 is LED
  TRISB = 0b11101001;
  //        11        ICSP pin
  //        ..1.....  MISO1
  //        ...0....  SCK1
  //        ....10..  EUSART debug I/O
  //        ......0.  no use
  //        .......1  INT interrupt in
  TRISC = 0b00000111;
  //        0.......  MOSI1
  //        .00.....  LCD control
  //        ...0....  no use
  //        .....1..  /SS2
  //        ......1.  MISO2
  //        .......1  SCK2
  LED = 0; // LEDは「割り込み処理中」に点灯する。
  // オシロスコープ等で割り込み処理時間を確認するため
  LCDCS = 1; // SPI関連信号の初期化
  LCDRS = 1;
  SCK = 0;
  iniz_spi1(); // TFT LCD module interface
// TMR0 - interval interrupt
  TMR0 = 0;
  T0CON = 0x00;
  T0CONbits.T08BIT = 1; // 8bit mode
  T0CONbits.T0PS = 0b111; // /256 pre-scale
                          // 8.192MHz / 4 / 256 /256 = 31.25(Hz)
  T0CONbits.TMR0ON = 1;
  INTCON2bits.T0IP = 0; // low priority
  INTCONbits.T0IE = 1;

// TMR1 - no interrupt, AC100V line period timing count
  TMR1 = 0; // TMR1はフリーランで動作させ、INT0割り込みでタイミングを測定する
  T1CON = 0x00;
  T1CONbits.T1CKPS = 0b01; // 1/2 pre-scale 2.048MHz counter
  T1CONbits.RD16 = 1; // 16bit read enable
  plc_sum = 0; // for averageing
  plc_n = 0, plc_avg = 0, plc_max = 0, plc_min = 0xFF;
  T1CONbits.TMR1ON = 1; // start timer

// INT0 interrupt - AC100V line period signal
  INTCONbits.INT0E = 1; // INT0割り込みはACラインのタイミングが入力されている

// TMR2 ADC trigger interrupt - 16KHz
  TMR2 = 0; // TMR2で内蔵ADコンバータのタイミングを生成
  PR2 = 255;
  T2CON = 0x00;
  pwf = 0;
  IPR1bits.TMR2IP = 1; // high priority
  T2CONbits.TMR2ON = 0; // not start now. - 内蔵ADCを使うときだけ動作させる

// SPI2 - MSSP2 - PCM1808からのデータをSPI信号として取得
  RPINR21 = 12; // MISO2 ; RC1/RP12 リマッパブルピンの設定
  RPINR22 = 11; // SCK2 : RC0/RP11
  RPINR23 = 13; // SS2 : RC2:RP13
  SSP2STAT = 0x00; 
  SSP2CON1 = 0x00;
  SSP2CON1bits.SSPM = 0b0100; // Slave mode, SS2 control enable
  IPR3bits.SSP2IP = 1; // high priority
  // MSSP2はPCM1808を使うときだけ動作させる

// interrupt enable
  RCONbits.IPEN = 1; // 優先度付き割り込みを使用
  INTCONbits.GIEH = 1; // start interrupt recept 割り込みを許可
  INTCONbits.GIEL = 1;
}

// PIC ADC data collection PIC内蔵ADCを使って波形を取得する
void
pic_adc_start( void )
{
  pwf = 0; // ポインタをクリア
  IPR1bits.TMR2IP = 1;  // TMR2割り込みを高優先度に
  TMR2 = 0;
  PR2 = 255; // 周期を16KSPSにする
  ADCON0 = 0b11000001; // ADコンバータ設定
  //         11...... // VREF is AN2/AN3
  //         ..0000.. // input channel is AN0
  //         ......0. // GO/DONE bit
  //         .......1 // ADC ON
  ADCON1 = 0b00001000;
  //         0....... // left justified
  //         .0...... // normal conversion
  //         ..001,,, // Acquisition time 2TAD
  //         .....000 // ADC clock is FOSC/2
  IPR1bits.ADIP = 1;  // high priority ADコンバータの割り込みも高優先度
  PIE1bits.ADIE = 1;  // interrupt enable 割り込み許可
  PIE1bits.TMR2IE = 1;
  T2CONbits.TMR2ON = 1; // start data collection TMR2をスタート
}

void
pic_waveform_sampling( void )
{
  while( !PORTBbits.RB0 ) ; // INT0割り込みでタイミング合わせ
  while( PORTBbits.RB0 ) ;
  pic_adc_start(); // 波形取得スタート
  while( T2CONbits.TMR2ON ) ; // 波形メモリがいっぱいになったら終了
  adc = ADCPIC; // 内蔵ADCでデータを取得したことを示すフラグ
}

// PCM1808で波形を取得する
void
pcm1808_adc_start( void )
{
  pwf = 0; // ポインタをクリア
  spi2bytecount = 0; // 当初4バイト取得しようとしていたときの残渣
  IPR3bits.SSP2IP = 1; // high priority
  SSP2CON1bits.SSPEN = 1; // MSSP2をスタート
  PIE3bits.SSP2IE = 1; // interrupt enable - start
}

void
pcm1808_waveform_sampling( void )
{
  int i;

  while( !PORTBbits.RB0 ) ; // INT0割り込みでタイミング合わせ
  while( PORTBbits.RB0 ) ;
  pcm1808_adc_start(); // 波形取得スタート
  while( PIE3bits.SSP2IE ) ; // 波形メモリがいっぱいになったら終了
// PCM1808 data : signed integer -> unsigned integer(offset binary)
　PCM1808の出力は符号付整数なのでPIC内蔵ADCに合わせてオフセットバイナリにする
  for( i = 0; i < WAVEFORM_BUFFER_SIZE; i ++ ) waveform[i] += 0x8000;
  adc = ADCPCM1808; // PCM1808でデータを取得したことを示すフラグ
}

// オフセットバイナリを扱うための平均値計算
long 
waveform_average( void )
{
  int i;
  long a;

  a = 0;
  for( i = 0; i < WAVEFORM_BUFFER_SIZE; i ++ ) a += waveform[i];
  return ( a / WAVEFORM_BUFFER_SIZE );
}

// MSP3520/21 flame draw : waveform
void
waveform_flame( void )
{
// draw flame
  vertical_line_ili9488( 240, 160, 160, C_WHITE );
  vertical_line_ili9488( 479, 160, 160, C_WHITE );
  horizontal_line_ili9488( 240, 160, 240, C_WHITE );
  horizontal_line_ili9488( 240, 319, 240, C_WHITE );
  vertical_line_ili9488( 280, 160, 160, C_GRAY );
  horizontal_line_ili9488( 280, 240, 200, C_GRAY );
  horizontal_line_ili9488( 280, 280, 200, C_GRAY );
  horizontal_line_ili9488( 280, 200, 200, C_GRAY );

  locate_ili9488( 245, 272 );
  puts_ili9488( "+100" );
  locate_ili9488( 245, 192 );
  puts_ili9488( "-100" );
}

// redraw waveform
void
waveform_draw( unsigned short int c )
{
  int i, x, y;
  long int a;

  for( x = 0; x < 198; x ++ ) {
    vertical_line_ili9488( 281 + x, 161, 158, C_BLACK );
    pset_ili9488( 281 + x, 200, C_GRAY );
    pset_ili9488( 281 + x, 240, C_GRAY );
    pset_ili9488( 281 + x, 280, C_GRAY );
    a = 0;
    for( i = 0; i < 4; i ++ ) a += ( long )waveform[ x * 4 + i ];
    y = 160 + ( int )( a / WAVEFORM_BUFFER_SIZE );
    pset_ili9488( 281 + x, y, c );
    display_on_ili9488();
  }
}

// overwrite waveform
void
waveform_draw2( unsigned short int c )
{
  int i, x, y;
  long int a;

  for( x = 0; x < 198; x ++ ) {
    a = 0;
    for( i = 0; i < 4; i ++ ) a += ( long )waveform[ x * 4 + i ];
    y = 160 + ( int )( a / WAVEFORM_BUFFER_SIZE );
    pset_ili9488( 281 + x, y, c );
    display_on_ili9488();
  }
}

int fgraph_x, vgraph_x;  // frequency / voltage graph X posirion

// MSP3520/21 flame draw : frequency
void
freq_graph_flame( void )
{
  char s[10];

// draw flame
  vertical_line_ili9488( 240, 0, 160, C_WHITE );
  vertical_line_ili9488( 479, 0, 160, C_WHITE );
  horizontal_line_ili9488( 240, 0, 240, C_WHITE );
  horizontal_line_ili9488( 240, 159, 240, C_WHITE );
  vertical_line_ili9488( 280, 0, 160, C_GRAY );
  horizontal_line_ili9488( 280, 80, 200, C_GRAY );
  horizontal_line_ili9488( 280, 120, 200, C_GRAY );
  horizontal_line_ili9488( 280, 40, 200, C_GRAY );
  locate_ili9488( 260, 72 );
  sprintf( s, "%d", freq_resion);
  puts_ili9488( s );
  locate_ili9488( 260, 112 );
  sprintf( s, "%d", freq_resion + 1 );
  puts_ili9488( s );
  locate_ili9488( 260, 32 );
  sprintf( s, "%d", freq_resion - 1 );
  puts_ili9488( s );  
  fgraph_x = 0;
}

// frequency graph
void
freq_graph_draw( unsigned short int c )
{
  unsigned long int f;
  int df;

  f = FINSTR * 1000 / plc_avg;
  df = f - freq_resion * 1000;
  vertical_line_ili9488( 281 + fgraph_x, 1, 158, C_BLACK );
  pset_ili9488( 281 + fgraph_x, 120, C_GRAY );
  pset_ili9488( 281 + fgraph_x, 80, C_GRAY );
  pset_ili9488( 281 + fgraph_x, 40, C_GRAY );
  pset_ili9488( 281 + fgraph_x, 80 + df / 25, c );
  display_on_ili9488();
  fgraph_x ++;
  if( 198 <= fgraph_x ) fgraph_x = 0;
}

// MSP3520/21 flame draw : voltage
void
voltage_graph_flame( void )
{
// draw flame
  vertical_line_ili9488( 0, 0, 160, C_WHITE );
  vertical_line_ili9488( 239, 0, 160, C_WHITE );
  horizontal_line_ili9488( 0, 0, 240, C_WHITE );
  horizontal_line_ili9488( 0, 159, 240, C_WHITE );
  vertical_line_ili9488( 40, 0, 160, C_GRAY );
  horizontal_line_ili9488( 40, 80, 200, C_GRAY );
  horizontal_line_ili9488( 40, 120, 200, C_GRAY );
  horizontal_line_ili9488( 40, 40, 200, C_GRAY );
  locate_ili9488( 12, 72 );
  puts_ili9488( "100" );
  locate_ili9488( 12, 112 );
  puts_ili9488( "105" );
  locate_ili9488( 20, 32 );
  puts_ili9488( "95" );  
  vgraph_x = 0;
}

// voltage graph
void
voltage_graph_draw( double v, unsigned short int c )
{
  int y;

  y = 80 + ( int )( ( v - 100.0 ) * 8.0 );
  vertical_line_ili9488( 41 + vgraph_x, 1, 158, C_BLACK );
  pset_ili9488( 41 + vgraph_x, 120, C_GRAY );
  pset_ili9488( 41 + vgraph_x, 80, C_GRAY );
  pset_ili9488( 41 + vgraph_x, 40, C_GRAY );
  pset_ili9488( 41 + vgraph_x, y, c );
  display_on_ili9488();
  vgraph_x ++;
  if( 198 <= vgraph_x ) vgraph_x = 0;
}

// root mean square 
// 実効値の計算 2乗して総和をとり平方根、のちに係数をかけて値を得る
double
calc_rms( void )
{
  int i;
  long int avg;
  double a, b;

// calc average( center voltage )
  avg = waveform_average();
// calc square sum 計算は浮動小数点で行っている
  a = 0.0;
  for( i = 0; i < WAVEFORM_BUFFER_SIZE; i ++ ) {
    b = ( double )( ( long )( waveform[i] - avg ) ) ;
    b /= 32768.0; // normalize
    a += b * b;
  }
  b = sqrt( a );
  if( ADCPIC == adc ) b *= COEF_V_PIC;
  else if( ADCPCM1808 == adc) b *= COEF_V_PCM1808;
  return b;
}

// integer sine / cosine 整数型の三角関数
//  a = angle 0..2047(2pi) 角度は0～2047で扱う
//  return: plus/minus 32767
int
i_sine( int a )
{

  a &= ( SINTABLE_SIZE - 1 );
  return sintable[ a ]; // フラッシュプログラムメモリ上のテーブルを参照
}

int
i_cosine( int a )
{
  if( ( SINTABLE_SIZE / 4 * 3 ) >= a ) return i_sine( a + SINTABLE_SIZE / 4 );
  else                                 return i_sine( a - ( SINTABLE_SIZE / 4 * 3 ) );
}

struct complex_l_type { // 複素数型を用意しておく
  long int real, imag;
};

// calucrate amplitude and phase 振幅と位相を計算する - フーリエ変換の基礎
struct complex_l_type
calc_ap( int freq )
{
  struct complex_l_type r;  // result
  unsigned long int p;  // phase
  int i, s, c, q;
  long int av;
//  char d[20];

  r.real = 0; r.imag = 0;
  p = 0;
  av = waveform_average(); 
  p = 0;
  for( i = 0; i < WAVEFORM_BUFFER_SIZE; i ++ ) {
    q = ( int )( p / 125 );
    s = i_sine( q ); // 実数部にはサインを
    c = i_cosine( q ); // 虚数部にはコサインを掛けて総和を得る
    r.real = r.real + ( long )s * ( ( long ) waveform[ i ] - av ) / 2048L;
    r.imag = r.imag + ( long )c * ( ( long ) waveform[ i ] - av ) / 2048L;
    p += ( long )freq * 16;
  }
  return r;
}

// ADコンバータで取得した波形データから
// （位相を合わせた）サイン波形を引いた残りが歪・ノイズ成分
void
subtract_sinwave( struct  complex_l_type ap, int freq )
{
  int i, q;
  long int av, p, r, j;

  av = waveform_average();
  ap.real /= 12800L;
  ap.imag /= 12800L;
  p = 0;
  for( i = 0;i < WAVEFORM_BUFFER_SIZE; i ++ ) {
    q = ( int )( p / 125 );
    j= ( ( long )i_sine( q ) * ap.real / 32768 +
         ( long )i_cosine( q ) * ap.imag / 32768 );
    waveform[i] = waveform[i] - j;
    p += ( long )freq * 16;
  }
}

void 
main(void) 
{
  int i, x;
  unsigned long int f;
  double v, thd;
  struct  complex_l_type a;
  char s[20];

// initialize 初期化
  iniz();
  iniz_ili9488();
  CLS();
// wait for PLC average stable 周波数取得が安定してできるまで待機
  locate_ili9488( 0, 300 );
  puts_ili9488( "Check Power line cycle..." );
  wait_ms( 1000 );
 // check Frequency resion(50Hz or 60Hz) 50Hzか60Hzかの判断 
  f = FINSTR / ( unsigned long )plc_avg;
  if( ( F50Hz + F60Hz ) / 2 < f ) freq_resion = F60Hz;
                               else freq_resion = F50Hz;
  sprintf( s, "%d Hz", freq_resion );
  puts_ili9488( s );
  wait_ms( 1000 );
  CLS();
// draw flame
  freq_graph_flame();
  voltage_graph_flame();
  waveform_flame();
// main loop
  while( 1 ) {
    WAVEFORM_SAMPLING();  // get waveform data
    v = calc_rms(); // rms voltage
    voltage_graph_draw( v, C_ORANGE );
    sprintf( s, "%3.2f V", v );
    if( 7 == strlen( s ) ) { s[7] = ' '; s[8] = '\0'; }
    locate_ili9488( 20, 260 );
    puts_symbol_ili9488( 3, s );
    waveform_draw( C_YELLOW ); // power line waveform
    a = calc_ap( freq_resion ); // THD waveform
    subtract_sinwave( a , freq_resion );
    waveform_draw2( C_RED );
    freq_graph_draw( C_GREEN );  // frequency graph
    f = FINSTR * 1000 / plc_avg; // frequency
    sprintf( s, "%2.2f Hz", ( double )f / 1000.0 );
    locate_ili9488( 20, 210 );
    puts_symbol_ili9488( 3, s );
    thd = calc_rms(); // THD %
    sprintf( s, "THD = %3.2f", thd / v * 100.0 );
    locate_ili9488( 20, 170 );
    puts_symbol_ili9488( 2, s );
    put_symbol_ili9488( '%', 2 );
  }
  return;
}
